/*****************************************************************************
 *
 * ORM128DS.cpp - Library for using OR-AVR-M128-DS methods.
 *	
 * Written by Vladimir Ermakov, Andrey Demenev, Botov Anton.
 * Copyright (c) 2010 Open Robotics project. For more information, see
 *
 *   http://open-robotics.ru/
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/atomic.h>
#include "ORM128DS.h"

#define debug(...)

#define DIR_PORT   PORTC
#define DIR_DDR    DDRC
#define DIR_MASK0  _BV(PC7)
#define DIR_MASK1  _BV(PC6)
#define DIR_MASK   (DIR_MASK0|DIR_MASK1)

#define PWM_OC0    OCR1A
#define PWM_OC1    OCR1B
#define PWM_PORT   PORTB
#define PWM_DDR    DDRB
#define PWM_MASK0  _BV(PB5) // OC1A
#define PWM_MASK1  _BV(PB6) // OC1B
#define PWM_MASK   (PWM_MASK0|PWM_MASK1) // avr/iom128.h OC1A, OC1B

// Timer1 OC1A, OC1B config
#define COM_PWM0           ((1<<COM1A1)|(0<<COM1A0))
#define COM_PWM1           ((1<<COM1B1)|(0<<COM1B0))
#define ATTACH_PWM_PIN(ch)  TCCR1A |= COM_PWM##ch
#define DETACH_PWM_PIN(ch)  TCCR1A &= ~COM_PWM##ch

/** Set motor direction
 * @param[in] ch channel [0, 1]
 * @param[in] value firward/backward [false/true]
 */
#define or_m128ds_motor_lld_set_direction(ch, value) \
	do {                                   \
		if (value)                         \
			DIR_PORT |= DIR_MASK##ch;      \
		else                               \
			DIR_PORT &= ~DIR_MASK##ch;     \
	} while(0)

#define or_m128ds_motor_lld_set_pwm(ch, value)   \
	do {                               \
		if (value) {                   \
			ATTACH_PWM_PIN(ch);        \
		} else {                       \
			DETACH_PWM_PIN(ch);        \
			PWM_PORT &= ~PWM_MASK##ch; \
		}                              \
		PWM_OC##ch = (uint16_t) value; \
	} while(0)

#define FF_CPU 7372800

#define US2CLOCK(us) (((uint32_t)(us) * (uint32_t)(FF_CPU / 8000000.0 * 0x10000UL)) >> 16)

#define process_timer(OCRX, TCCRX, block, FOC_MASK) {	\
	OCRX += *(table_ptr[block]);						\
	if (table_ptr[block] == &(calc_ocr[block][8]) ) {	\
		table_ptr[block] = &(calc_ocr[block][0]);		\
		TCCRX |= FOC_MASK;								\
		TCCRX |= FOC_MASK;								\
	} else {											\
		table_ptr[block]++;								\
	}													\
	TCCRX |= FOC_MASK;									\
}

static uint16_t servo_pos[16];

static uint8_t PROGMEM pin_map[16] = {
	7, 3, 2, 6, 5, 1, 0, 4,
	4, 0, 1, 5, 6, 2, 3, 7,
};

static uint16_t calc_ocr[2][9] = {
	{
		US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500),
		US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500),
		US2CLOCK(20500 - 8 * 1500),
	},
	{
		US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500),
		US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500), US2CLOCK(1500),
		US2CLOCK(20500 - 8 * 1500),
	},
};

static uint16_t* table_ptr[2] = {
	calc_ocr[0],
	calc_ocr[1],
};


ISR(SIG_OUTPUT_COMPARE3A) {
	process_timer(OCR3A, TCCR3C, 0, (1 << FOC3A));
}

ISR(SIG_OUTPUT_COMPARE3C) {
	process_timer(OCR3C, TCCR3C, 1, (1 << FOC3C));
}


uint16_t or_m128ds_servo_get_position(uint8_t n)
{
	if (n > SERVO_CHMAX_DS)
		return 0;
	return servo_pos[n];
}

void or_m128ds_servo_set_position(uint8_t n, uint16_t pos)
{
	if (n > SERVO_CHMAX_DS)
		return;

	if (pos < 500)
		pos = 500;
	else if (pos > 2500)
		pos = 2500;

	uint8_t idx = pgm_read_byte(pin_map+n);
	uint8_t block = n >> 3;
	uint16_t uspos = US2CLOCK(pos);

	uint16_t* pos_p = servo_pos+n;
	uint16_t* pause_p = calc_ocr[block]+8;
	uint16_t* pulse_p = calc_ocr[block]+idx;

	uint16_t pause = *pause_p;
	uint16_t pulse = *pulse_p;

	pause += pulse;
	pause -= uspos;
	
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		*pause_p = pause;
		*pulse_p = uspos;
		*pos_p = pos;
	}
}

void or_m128ds_motor_set_speed(int16_t speed_left,int16_t speed_right) //speed=-255..255
{
	if(speed_left>=0){
		or_m128ds_motor_lld_set_direction(0, 0);
		or_m128ds_motor_lld_set_pwm(0, speed_left);
	}else{
		or_m128ds_motor_lld_set_direction(0, 1);
		or_m128ds_motor_lld_set_pwm(0, -speed_left);

	};
	if(speed_right>=0){
		or_m128ds_motor_lld_set_direction(1, 0);
		or_m128ds_motor_lld_set_pwm(1, speed_right);
	}else{
		or_m128ds_motor_lld_set_direction(1, 1);
		or_m128ds_motor_lld_set_pwm(1, -speed_right);

	};
}

void or_m128ds_motor_lld_init(void)
{
	// init i/o lines
	DIR_PORT &= ~DIR_MASK;
	DIR_DDR |= DIR_MASK;
	PWM_PORT &= ~PWM_MASK;
	PWM_DDR |= PWM_MASK;

	// clear pwm registers
	PWM_OC0 = 0;
	PWM_OC1 = 0;

	// Timer/Counter-1 init
	//
	// COM1x1:COM1x0 = 1:0 => OC1x connected to output pin, reset on compare
	// FOC1A:FOC1B = 0:0 => Don't force output compare event
	// WGM13:WGM12:WGM11:WGM10 = 0:1:0:1 => Fast PWM 8-bit
	// ICNC1 = 0 => Input Capture Noise Canceller disabled (because not needed)
	// ICES1 = 0 => Input Capture Edge Select set to "falling" mode (no matter)
	// CS12:CS11:CS10 = 0:1:0 => Timer "on", 1/8 prescaling (1 clock every 8 cpu tact)	// /8
	TCNT1 = 0;
	TCCR1A = (0<<WGM11)|(1<<WGM10);
	TCCR1B = (0<<WGM13)|(1<<WGM12)|(0<<CS12)|(1<<CS11)|(0<<CS10);
}


void or_m128ds_init(void)
{
	DDRE = _BV(2) | _BV(3) | _BV(5);

	PORTE |= _BV(2);
	
	OCR3A = 2000;
	OCR3C = 2000;

	TCCR3A = _BV(COM3A0) | _BV(COM3C0);
	TCCR3B = _BV(CS31);

	ETIMSK |= _BV(OCIE3A) | _BV(OCIE3C);

	PORTE &= ~_BV(2);

	or_m128ds_motor_lld_init();

	sei();
}


